"""
Streams module. Handles reading and writing to streams.
Backed by Redis Streams
"""
from abc import ABC
import os
import socket
from datetime import datetime, timezone
from typing import Dict, List, Any, Union
from dataclasses import dataclass
from importlib import import_module

from hopeit.app.config import Compression, Serialization
from hopeit.dataobjects import EventPayload
from hopeit.server.config import AuthType, StreamsConfig
from hopeit.server.logger import engine_logger, extra_logger

logger = engine_logger()
extra = extra_logger()

__all__ = ['StreamEvent',
           'StreamManager',
           'stream_auth_info',
           'StreamOSError']


@dataclass
class StreamEvent:
    msg_internal_id: bytes
    queue: str
    payload: EventPayload
    track_ids: Dict[str, str]
    auth_info: Dict[str, Any]


class StreamOSError(Exception):
    pass


class StreamConfigError(Exception):
    pass


def stream_auth_info(stream_event: StreamEvent):
    return {
        **stream_event.auth_info,
        'auth_type': AuthType(stream_event.auth_info.get('auth_type', 'Unsecured')),
    }


class StreamManager(ABC):
    """
    Base class to imeplement stream management of a Hopeit App
    """
    @staticmethod
    def create(config: StreamsConfig):
        sm_comps = config.stream_manager.split('.')
        module_name, impl_name = '.'.join(sm_comps[:-1]), sm_comps[-1]
        logger.info(__name__, f"Importing StreamManager module: {module_name} implementation: {impl_name}...")
        module = import_module(module_name)
        impl = getattr(module, impl_name)
        logger.info(__name__, f"Creating {impl_name} with connection_str: {config.connection_str}...")
        return impl(address=config.connection_str)

    async def connect(self):
        """
        Connects to Redis using two connection pools, one to handle
        writing to stream and one for reading.
        """
        raise NotImplementedError()

    async def close(self):
        """
        Close connections to Redis
        """
        raise NotImplementedError()

    async def write_stream(self, *,
                           stream_name: str,
                           queue: str,
                           payload: EventPayload,
                           track_ids: Dict[str, str],
                           auth_info: Dict[str, Any],
                           compression: Compression,
                           serialization: Serialization,
                           target_max_len: int = 0) -> int:
        """
        Writes event to a Redis stream using XADD
        :param stream_name: stream name or key used by Redis
        :param payload: EventPayload, a special type of dataclass object decorated with `@dataobject`
        :param track_ids: dict with key and id values to track in stream event
        :param auth_info: dict with auth info to be tracked as part of stream event
        :param compression: Compression, supported compression algorithm from enum
        :param target_max_len: int, max_len to indicate approx. target collection size to Redis,
            default 0 will not send max_len to Redis.
        :return: number of successful written messages
        """
        raise NotImplementedError()

    async def ensure_consumer_group(self, *,
                                    stream_name: str,
                                    consumer_group: str):
        """
        Ensures a consumer_group exists for a given stream.
        If group does not exists, XGROUP_CREATE will be executed in Redis and consumer group
        created to consume event from beginning of stream (from id=0)
        If group already exists a message will be logged and no action performed.
        If stream does not exists and empty stream will be created.
        :param stream_name: str, stream name or key used by Redis
        :param consumer_group: str, consumer group passed to Redis
        """
        raise NotImplementedError()

    async def read_stream(self, *,
                          stream_name: str,
                          consumer_group: str,
                          datatypes: Dict[str, type],
                          track_headers: List[str],
                          offset: str,
                          batch_size: int,
                          timeout: int,
                          batch_interval: int) -> List[Union[StreamEvent, Exception]]:
        """
        Attempts reading streams using a consumer group,
        blocks for `timeout` seconds
        and yields asynchronously the deserialized objects gotten from the stream.
        In case timeout is reached, nothing is yielded
        and read_stream must be called again,
        usually in an infinite loop while app is running.
        :param stream_name: str, stream name or key used by Redis
        :param consumer_group: str, consumer group registered in Redis
        :param datatypes: Dict[str, type] supported datatypes name: type to be extracted from stream.
            Types need to support json deserialization using `@dataobject` annotation
        :param track_headers: list of headers/id fields to extract from message if available
        :param offset: str, last msg id consumed to resume from. Use'>' to consume unconsumed events,
            or '$' to consume upcoming events
        :param batch_size: max number of messages to process on each iteration
        :param timeout: time to block waiting for messages, in milliseconds
        :param batch_interval: int, time to sleep between requests to connection pool in case no
            messages are returned. In milliseconds. Used to prevent blocking the pool.
        :param compression: Compression, supported compression algorithm from enum
        :return: yields Tuples of message id (bytes) and deserialized DataObject
        """
        raise NotImplementedError()

    async def ack_read_stream(self, *,
                              stream_name: str,
                              consumer_group: str,
                              stream_event: StreamEvent):
        """
        Acknowledges a read message to Redis streams.
        Acknowledged messages are removed from a pending list by Redis.
        This method should be called for every message that is properly
        received and processed with no errors.
        With this mechanism, messages not acknowledged can be retried.
        :param stream_name: str, stream name or key used by Redis
        :param consumer_group: str, consumer group registered with Redis
        :param stream_event: StreamEvent, as provided by `read_stream(...)` method
        """
        raise NotImplementedError()

    @staticmethod
    def as_data_event(payload: EventPayload) -> EventPayload:
        """
        Checks payload for implementing `@dataobject` decorator.
        Raises NotImplementedError if payload does not implement `@dataobject`
        :param payload: dataclass object decorated with `@dataobject`
        :return: same payload as received
        """
        if not getattr(payload, '__data_object__', False):
            raise NotImplementedError(
                f"{type(payload)} must be decorated with `@dataobject` to be used in streams")
        return payload

    def _consumer_id(self) -> str:
        """
        Constructs a consumer id for this instance
        :return: str, concatenating current UTC ISO datetime, host name, process id
            and this StreamManager instance id
        """
        ts = datetime.now(tz=timezone.utc).isoformat()
        host = socket.gethostname()
        pid = os.getpid()
        mgr = id(self)
        return f"{ts}.{host}.{pid}.{mgr}"


class NoStreamManager(StreamManager):
    """
    Default noop implementation for StreamManager that will fail if no Stream Manager is specified in
    server config file. Applications using streams will fail to start a server if
    `stream_manager` is not properly specified in configuration file.
    """
    def __init__(self, *, address: str):
        raise StreamConfigError(
            "Cannot start engine StreamManager, "
            "a valid `stream_manager` implementation needs to be specified."
            "\n You need to install a proper Stream Manager."
            "\n i.e. to use Redis Streams: "
            "\n 1) Install `hopeit.engine` with redis-streams plugin: `pip install hopeit.engine[redis-streams]`"
            "\n 2) Add the following entry to server config json file under `streams` section:"
            '"stream_manager": "hopeit.redis_streams.RedisStreamManager", '
            "\nCheck Plugins/ -> Streams docs section for available additional options."
        )



"""
Json tools to serialized and deserialze data objects
"""

import warnings

from hopeit.dataobjects.payload import Payload


class Json(Payload):
    """
    Json convenience ser/deser functions for @dataobject decorated object (@see DataObject)
    """
    warnings.warn(
        "Usage of `Json` from `hopeit.engine.jsonify` is deprecated since 0.8.3 and will be removed "
        "in a future version. Use `Payload` from `hopeit.engine.payload` instead", DeprecationWarning
    )



"""
Webserver module based on aiohttp to handle web/api requests
"""
# flake8: noqa
# pylint: disable=wrong-import-position, wrong-import-order
from collections import namedtuple
import aiohttp

setattr(aiohttp.http, 'SERVER_SOFTWARE', '')

import argparse
import asyncio
import gc
import logging
import re
import sys
import uuid
from datetime import datetime, timezone
from functools import partial
from typing import (
    Any, Callable, Coroutine, Dict, List, Optional, Type, Tuple, Union
)

import aiohttp_cors  # type: ignore
from aiohttp import web
from aiohttp.web_response import Response
from aiohttp_cors import CorsConfig
from stringcase import snakecase, titlecase  # type: ignore

from hopeit.app.config import (
    AppConfig, EventDescriptor, EventPlugMode, EventSettings, EventType, parse_app_config_json
)
from hopeit.app.context import (
    EventContext, NoopMultiparReader, PostprocessHook, PreprocessHook
)
from hopeit.app.errors import BadRequest, Unauthorized
from hopeit.dataobjects import DataObject, EventPayload, EventPayloadType
from hopeit.dataobjects.payload import Payload
from hopeit.server import api, runtime
from hopeit.server.api import app_route_name
from hopeit.server.config import (
    AuthType, ServerConfig, parse_server_config_json
)
from hopeit.server.engine import AppEngine, Server
from hopeit.server.errors import ErrorInfo
from hopeit.server.events import get_event_settings
from hopeit.server.logger import (
    EngineLoggerWrapper, combined, engine_logger, extra_logger
)
from hopeit.server.metrics import metrics
from hopeit.server.names import route_name
from hopeit.server.steps import find_datatype_handler
from hopeit.toolkit import auth

__all__ = ['parse_args',
           'prepare_engine',
           'serve',
           'server_startup_hook',
           'app_startup_hook',
           'stream_startup_hook',
           'stop_server']

logger: EngineLoggerWrapper = logging.getLogger(__name__)  # type: ignore
extra = extra_logger()

ResponseType = Union[web.Response, web.FileResponse, web.StreamResponse]

# server = Server()
web_server = web.Application()
auth_info_default = {}


def prepare_engine(*, config_files: List[str], api_file: Optional[str], enabled_groups: List[str], start_streams: bool):
    """
    Load configuration files and add hooks to setup engine server and apps,
    start streams and services.
    """
    logger.info("Loading engine config file=%s...", config_files[0])  # type: ignore
    server_config: ServerConfig = _load_engine_config(config_files[0])

    # Add startup hook to start engine
    web_server.on_startup.append(
        partial(server_startup_hook, server_config)
    )
    if server_config.auth.domain:
        auth_info_default['domain'] = server_config.auth.domain
    if api_file is not None:
        api.load_api_file(api_file)
        api.register_server_config(server_config)

    apps_config = []
    for config_file in config_files[1:]:
        logger.info(__name__, f"Loading app config file={config_file}...")
        config = _load_app_config(config_file)
        config.server = server_config
        apps_config.append(config)

    # Register and add startup hooks to start configured apps
    api.register_apps(apps_config)
    api.enable_swagger(server_config, web_server)
    for config in apps_config:
        web_server.on_startup.append(
            partial(app_startup_hook, config, enabled_groups)
        )

    # Add hooks to start streams and service
    if start_streams:
        for config in apps_config:
            web_server.on_startup.append(
                partial(stream_startup_hook, config)
            )

    logger.debug(__name__, "Performing forced garbage collection...")
    gc.collect()


def serve(*, host: str, path: str, port: int):
    logger.info(__name__, f"Starting web server host: {host} port: {port} socket: {path}...")
    web.run_app(web_server, host=host, path=path, port=port)


def init_logger():
    global logger
    logger = engine_logger()


async def server_startup_hook(config: ServerConfig, *args, **kwargs):
    """
    Start engine engine
    """
    await runtime.server.start(config=config)
    init_logger()


async def stop_server():
    global web_server
    await runtime.server.stop()
    await web_server.shutdown()
    runtime.server = Server()
    web_server = web.Application()


async def app_startup_hook(config: AppConfig, enabled_groups: List[str], *args, **kwargs):
    """
    Start Hopeit app specified by config

    :param config: AppConfig, configuration for the app to start
    :param enabled_groups: list of event groups names to enable. If empty,
        all events will be enabled.
    """
    app_engine = await runtime.server.start_app(app_config=config, enabled_groups=enabled_groups)
    cors_origin = aiohttp_cors.setup(web_server, defaults={
        config.engine.cors_origin: aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    }) if config.engine.cors_origin else None

    _setup_app_event_routes(app_engine)
    for plugin in config.plugins:
        plugin_engine = runtime.server.app_engine(app_key=plugin.app_key())
        _setup_app_event_routes(app_engine, plugin_engine)
    if cors_origin:
        app = app_engine.app_config.app
        _enable_cors(route_name('api', app.name, app.version), cors_origin)


async def stream_startup_hook(app_config: AppConfig, *args, **kwargs):
    """
    Start all stream event types configured in app.

    :param app_key: already started app_key
    """
    app_engine = runtime.server.app_engines[app_config.app_key()]
    for event_name, event_info in app_engine.effective_events.items():
        if event_info.type == EventType.STREAM:
            assert event_info.read_stream
            logger.info(
                __name__, f"STREAM start event_name={event_name} read_stream={event_info.read_stream.name}")
            asyncio.create_task(app_engine.read_stream(event_name=event_name))
        elif event_info.type == EventType.SERVICE:
            logger.info(
                __name__, f"SERVICE start event_name={event_name}")
            asyncio.create_task(app_engine.service_loop(event_name=event_name))


def _effective_events(app_engine: AppEngine, plugin: Optional[AppEngine] = None):
    if plugin is None:
        return {
            k: v for k, v in app_engine.effective_events.items()
            if v.plug_mode == EventPlugMode.STANDALONE
        }
    return {
        k: v for k, v in plugin.effective_events.items()
        if v.plug_mode == EventPlugMode.ON_APP
    }


def _load_engine_config(path: str):
    """
    Load engine configuration from json file
    """
    with open(path, encoding="utf-8") as f:
        return parse_server_config_json(f.read())


def _load_app_config(path: str) -> AppConfig:
    """
    Load app configuration from json file
    """
    with open(path, encoding="utf-8") as f:
        return parse_app_config_json(f.read())


def _enable_cors(prefix: str, cors: CorsConfig):
    for route in web_server.router.routes():
        if route.resource and route.resource.canonical.startswith(prefix):
            cors.add(route)


def _setup_app_event_routes(app_engine: AppEngine,
                            plugin: Optional[AppEngine] = None):
    """
    Setup http routes for existing events in app,
    in existing web_server global instance.

    Supports:
    * GET requests with query params
    * POST requests with query params and payload sent in body
    * STREAM start/stop endpoints

    :param app_engine: AppEngine, initialized application engine
    :param cors: initialized CorsConfig object or None is CORS is disabled
    :param plugin: optional AppEngine, when the implementation of the route
        is handled by a plugin app, if not specified methods will be handled
        by same app_engine
    """
    for event_name, event_info in _effective_events(app_engine, plugin).items():
        if event_info.type == EventType.POST:
            web_server.add_routes([
                _create_post_event_route(
                    app_engine, plugin=plugin, event_name=event_name, event_info=event_info
                )
            ])
        elif event_info.type == EventType.GET:
            web_server.add_routes([
                _create_get_event_route(
                    app_engine, plugin=plugin, event_name=event_name, event_info=event_info
                )
            ])
        elif event_info.type == EventType.MULTIPART:
            web_server.add_routes([
                _create_multipart_event_route(
                    app_engine, plugin=plugin, event_name=event_name, event_info=event_info
                )
            ])
        elif event_info.type == EventType.STREAM and plugin is None:
            web_server.add_routes(
                _create_event_management_routes(
                    app_engine, event_name=event_name, event_info=event_info
                )
            )
        elif event_info.type == EventType.SERVICE and plugin is None:
            web_server.add_routes(
                _create_event_management_routes(
                    app_engine, event_name=event_name, event_info=event_info
                )
            )
        else:
            raise ValueError(f"Invalid event_type:{event_info.type} for event:{event_name}")


def _auth_types(app_engine: AppEngine, event_name: str):
    assert app_engine.app_config.server
    event_info = app_engine.app_config.events[event_name]
    if event_info.auth:
        return event_info.auth
    return app_engine.app_config.server.auth.default_auth_methods


def _create_post_event_route(
        app_engine: AppEngine, *,
        plugin: Optional[AppEngine] = None,
        event_name: str,
        event_info: EventDescriptor) -> web.RouteDef:
    """
    Creates route for handling POST event
    """
    datatype = find_datatype_handler(app_config=app_engine.app_config, event_name=event_name, event_info=event_info)
    route = app_route_name(app_engine.app_config.app, event_name=event_name,
                           plugin=None if plugin is None else plugin.app_config.app,
                           override_route_name=event_info.route)
    logger.info(__name__, f"POST path={route} input={str(datatype)}")
    impl = plugin if plugin else app_engine
    handler = partial(_handle_post_invocation, app_engine, impl,
                      event_name, datatype, _auth_types(impl, event_name))
    setattr(handler, '__closure__', None)
    setattr(handler, '__code__', _handle_post_invocation.__code__)
    api_handler = api.add_route('post', route, handler)
    return web.post(route, api_handler)


def _create_get_event_route(
        app_engine: AppEngine, *,
        plugin: Optional[AppEngine] = None,
        event_name: str,
        event_info: EventDescriptor) -> web.RouteDef:
    """
    Creates route for handling GET requests
    """
    route = app_route_name(app_engine.app_config.app, event_name=event_name,
                           plugin=None if plugin is None else plugin.app_config.app,
                           override_route_name=event_info.route)
    logger.info(__name__, f"GET path={route}")
    impl = plugin if plugin else app_engine
    handler = partial(_handle_get_invocation, app_engine, impl, event_name, _auth_types(impl, event_name))
    setattr(handler, '__closure__', None)
    setattr(handler, '__code__', _handle_get_invocation.__code__)
    api_handler = api.add_route('get', route, handler)
    return web.get(route, api_handler)


def _create_multipart_event_route(
        app_engine: AppEngine, *,
        plugin: Optional[AppEngine] = None,
        event_name: str,
        event_info: EventDescriptor) -> web.RouteDef:
    """
    Creates route for handling MULTIPART event
    """
    datatype = find_datatype_handler(app_config=app_engine.app_config, event_name=event_name, event_info=event_info)
    route = app_route_name(app_engine.app_config.app, event_name=event_name,
                           plugin=None if plugin is None else plugin.app_config.app,
                           override_route_name=event_info.route)
    logger.info(__name__, f"MULTIPART path={route} input={str(datatype)}")
    impl = plugin if plugin else app_engine
    handler = partial(_handle_multipart_invocation, app_engine, impl,
                      event_name, datatype, _auth_types(impl, event_name))
    setattr(handler, '__closure__', None)
    setattr(handler, '__code__', _handle_multipart_invocation.__code__)
    api_handler = api.add_route('post', route, handler)
    return web.post(route, api_handler)


def _create_event_management_routes(
        app_engine: AppEngine, *,
        event_name: str,
        event_info: EventDescriptor) -> List[web.RouteDef]:
    """
    Create routes to start and stop processing of STREAM events
    """
    evt = event_name.replace('.', '/').replace('$', '/')
    base_route = app_route_name(app_engine.app_config.app, event_name=evt,
                                prefix='mgmt', override_route_name=event_info.route)
    logger.info(__name__, f"{event_info.type.value.upper()} path={base_route}/[start|stop]")

    handler: Optional[partial[Coroutine[Any, Any, Response]]] = None
    if event_info.type == EventType.STREAM:
        handler = partial(_handle_stream_start_invocation, app_engine, event_name)
    elif event_info.type == EventType.SERVICE:
        handler = partial(_handle_service_start_invocation, app_engine, event_name)
    assert handler is not None, f"No handler for event={event_name} type={event_info.type}"
    return [
        web.get(base_route + '/start', handler),
        web.get(
            base_route + '/stop',
            partial(_handle_event_stop_invocation, app_engine, event_name)
        )
    ]


def _response(*, track_ids: Dict[str, str], key: str, payload: EventPayload, hook: PostprocessHook) -> ResponseType:
    """
    Creates a web response object from a given payload (body), header track ids
    and applies a postprocess hook
    """
    response: ResponseType
    headers = {
        **hook.headers,
        **{f"X-{re.sub(' ', '-', titlecase(k))}": v for k, v in track_ids.items()}
    }
    if hook.file_response is not None:
        response = web.FileResponse(
            path=hook.file_response,
            headers={'Content-Type': hook.content_type, **headers}
        )
    elif hook.stream_response is not None:
        response = hook.stream_response.resp
    else:
        serializer: Callable[..., str] = CONTENT_TYPE_BODY_SER.get(
            hook.content_type, _text_response
        )
        body = serializer(payload, key=key)
        response = web.Response(
            body=body,
            headers=headers,
            content_type=hook.content_type
        )
        for name, cookie in hook.cookies.items():
            value, args, kwargs = cookie
            response.set_cookie(name, value, *args, **kwargs)
        for name, args, kwargs in hook.del_cookies:
            response.del_cookie(name, *args, **kwargs)
    if hook.status:
        response.set_status(hook.status)
    return response


def _response_info(response: ResponseType):
    return extra(prefix='response.', status=str(response.status))


def _track_ids(request: web.Request) -> Dict[str, str]:
    return {
        'track.operation_id': str(uuid.uuid4()),
        'track.request_id': str(uuid.uuid4()),
        'track.request_ts': datetime.now(tz=timezone.utc).isoformat(),
        **{
            "track." + snakecase(k[8:].lower()): v
            for k, v in request.headers.items() if k.lower().startswith('x-track-')
        }
    }


def _failed_response(context: Optional[EventContext],
                     e: Exception) -> web.Response:
    if context:
        logger.error(context, e)
        logger.failed(context)
    else:
        logger.error(__name__, e)
    info = ErrorInfo.from_exception(e)
    return web.Response(
        status=500,
        body=Payload.to_json(info)
    )


def _ignored_response(context: Optional[EventContext],
                      status: int,
                      e: BaseException) -> web.Response:
    if context:
        logger.error(context, e)
        logger.ignored(context)
    else:
        logger.error(__name__, e)
    info = ErrorInfo.from_exception(e)
    return web.Response(
        status=status,
        body=Payload.to_json(info)
    )


def _request_start(app_engine: AppEngine,
                   plugin: AppEngine,
                   event_name: str,
                   event_settings: EventSettings,
                   request: web.Request) -> EventContext:
    """
    Extracts context and track information from a request and logs start of event
    """
    context = EventContext(
        app_config=app_engine.app_config,
        plugin_config=plugin.app_config,
        event_name=event_name,
        settings=event_settings,
        track_ids=_track_ids(request),
        auth_info=auth_info_default
    )
    logger.start(context)
    return context


def _extract_auth_header(request: web.Request, context: EventContext) -> Optional[str]:
    return request.headers.get("Authorization")


def _extract_refresh_cookie(request: web.Request, context: EventContext) -> Optional[str]:
    return request.cookies.get(f"{context.app_key}.refresh")


def _ignore_auth(request: web.Request, context: EventContext) -> str:
    return 'Unsecured -'


AUTH_HEADER_EXTRACTORS = {
    AuthType.BASIC: _extract_auth_header,
    AuthType.BEARER: _extract_auth_header,
    AuthType.REFRESH: _extract_refresh_cookie,
    AuthType.UNSECURED: _ignore_auth
}


def _extract_authorization(auth_methods: List[AuthType], request: web.Request, context: EventContext):
    for auth_type in auth_methods:
        auth_header = AUTH_HEADER_EXTRACTORS[auth_type](request, context)
        if auth_header is not None:
            return auth_header
    return 'Unsecured -'


def _validate_authorization(app_config: AppConfig,
                            context: EventContext,
                            auth_types: List[AuthType],
                            request: web.Request):
    """
    Validates Authorization header from request to provide valid credentials
    for the methods supported in event configuration.

    :raise `Unauthorized` if authorization is not valid
    """
    auth_methods = context.event_info.auth
    if (len(auth_methods) == 0) and (app_config.server is not None):
        auth_methods = app_config.server.auth.default_auth_methods
    auth_header = _extract_authorization(auth_methods, request, context)

    try:
        method, data = auth_header.split(" ")
    except ValueError as e:
        raise BadRequest("Malformed Authorization") from e

    context.auth_info['allowed'] = False
    for auth_type in auth_types:
        if method.upper() == auth_type.name.upper():
            auth.validate_auth_method(auth_type, data, context)
            if context.auth_info.get('allowed'):
                return None
    raise Unauthorized(method)


def _application_json_response(result: DataObject, key: str, *args, **kwargs) -> str:
    return Payload.to_json(result, key=key)


def _text_response(result: str, *args, **kwargs) -> str:
    return str(result)


CONTENT_TYPE_BODY_SER: Dict[str, Callable[..., str]] = {
    'application/json': _application_json_response,
    'text/html': _text_response,
    'text/plain': _text_response
}


async def _request_execute(
        app_engine: AppEngine,
        event_name: str,
        context: EventContext,
        query_args: Dict[str, Any],
        payload: Optional[EventPayloadType],
        preprocess_hook: PreprocessHook,
        request: web.Request) -> ResponseType:
    """
    Executes request using engine event handler
    """
    response_hook = PostprocessHook(request)
    result = await app_engine.preprocess(
        context=context, query_args=query_args, payload=payload, request=preprocess_hook)
    if (preprocess_hook.status is None) or (preprocess_hook.status == 200):
        result = await app_engine.execute(context=context, query_args=query_args, payload=result)
        result = await app_engine.postprocess(context=context, payload=result, response=response_hook)
    else:
        response_hook.set_status(preprocess_hook.status)
    response = _response(
        track_ids=context.track_ids,
        key=event_name,
        payload=result,
        hook=response_hook
    )
    logger.done(context, extra=combined(
        _response_info(response), metrics(context)
    ))
    return response


async def _request_process_payload(
        context: EventContext,
        datatype: Optional[Type[EventPayloadType]],
        request: web.Request) -> Tuple[Optional[EventPayloadType], Optional[bytes]]:
    """
    Extract payload from request.
    Returns payload if parsing succeeded. Raises BadRequest if payload fails to parse
    """
    try:
        payload_raw = await request.read()
        if datatype is not None:
            return Payload.from_json(payload_raw, datatype), payload_raw  # type: ignore
        return None, payload_raw
    except ValueError as e:
        logger.error(context, e)
        raise BadRequest(e) from e


async def _handle_post_invocation(
        app_engine: AppEngine,
        impl: AppEngine,
        event_name: str,
        datatype: Optional[Type[DataObject]],
        auth_types: List[AuthType],
        request: web.Request) -> ResponseType:
    """
    Handler to execute POST calls
    """
    context = None
    try:
        event_settings = get_event_settings(app_engine.settings, event_name)
        context = _request_start(app_engine, impl, event_name, event_settings, request)
        query_args = dict(request.query)
        _validate_authorization(app_engine.app_config, context, auth_types, request)
        payload, payload_raw = await _request_process_payload(context, datatype, request)
        hook: PreprocessHook[NoopMultiparReader] = PreprocessHook(headers=request.headers, payload_raw=payload_raw)
        return await _request_execute(
            impl, event_name, context, query_args, payload, preprocess_hook=hook, request=request
        )
    except Unauthorized as e:
        return _ignored_response(context, 401, e)
    except BadRequest as e:
        return _ignored_response(context, 400, e)
    except Exception as e:  # pylint: disable=broad-except
        return _failed_response(context, e)


async def _handle_get_invocation(
        app_engine: AppEngine,
        impl: AppEngine,
        event_name: str,
        auth_types: List[AuthType],
        request: web.Request) -> ResponseType:
    """
    Handler to execute GET calls
    """
    context = None
    try:
        event_settings = get_event_settings(app_engine.settings, event_name)
        context = _request_start(app_engine, impl, event_name, event_settings, request)
        _validate_authorization(app_engine.app_config, context, auth_types, request)
        query_args = dict(request.query)
        payload = query_args.get('payload')
        if payload is not None:
            del query_args['payload']
        hook: PreprocessHook[NoopMultiparReader] = PreprocessHook(headers=request.headers)
        return await _request_execute(
            impl, event_name, context,
            query_args, payload=payload,
            preprocess_hook=hook,
            request=request
        )
    except Unauthorized as e:
        return _ignored_response(context, 401, e)
    except BadRequest as e:
        return _ignored_response(context, 400, e)
    except Exception as e:  # pylint: disable=broad-except
        return _failed_response(context, e)


async def _handle_multipart_invocation(
        app_engine: AppEngine,
        impl: AppEngine,
        event_name: str,
        datatype: Optional[Type[DataObject]],
        auth_types: List[AuthType],
        request: web.Request) -> ResponseType:
    """
    Handler to execute POST calls
    """
    context = None
    try:
        event_settings = get_event_settings(app_engine.settings, event_name)
        context = _request_start(app_engine, impl, event_name, event_settings, request)
        query_args = dict(request.query)
        _validate_authorization(app_engine.app_config, context, auth_types, request)
        hook = PreprocessHook(                                                   # type: ignore
            headers=request.headers, multipart_reader=await request.multipart()  # type: ignore
        )
        return await _request_execute(
            impl, event_name, context,
            query_args, payload=None,
            preprocess_hook=hook,
            request=request
        )
    except Unauthorized as e:
        return _ignored_response(context, 401, e)
    except BadRequest as e:
        return _ignored_response(context, 400, e)
    except Exception as e:  # pylint: disable=broad-except
        return _failed_response(context, e)


async def _handle_stream_start_invocation(
        app_engine: AppEngine,
        event_name: str,
        request: web.Request) -> web.Response:
    """
    Handles call to stream processing event `start` endpoint,
    spawning an async job that listens continuosly to event streams
    in the background.
    """
    assert request
    if app_engine.is_running(event_name):
        return web.Response(status=500, body=f"Stream already running: {event_name}")
    asyncio.create_task(app_engine.read_stream(event_name=event_name))
    return web.Response()


async def _handle_service_start_invocation(
        app_engine: AppEngine,
        event_name: str,
        request: web.Request) -> web.Response:
    """
    Handles call to service event `start` endpoint,
    spawning an async job that listens continuosly __service__
    generator in the background
    """
    assert request
    if app_engine.is_running(event_name):
        return web.Response(status=500, body=f"Service already running: {event_name}")
    asyncio.create_task(app_engine.service_loop(event_name=event_name))
    return web.Response()


async def _handle_event_stop_invocation(
        app_engine: AppEngine,
        event_name: str,
        request: web.Request) -> web.Response:
    """
    Signals engine for stopping an event.
    Used to stop reading stream processing events.
    """
    try:
        assert request
        await app_engine.stop_event(event_name)
        logger.info(__name__, f"Event stop signaled event_name={event_name}...")
        return web.Response()
    except RuntimeError as e:
        return web.Response(status=500, body=str(e))


ParsedArgs = namedtuple("ParsedArgs", ["host", "port", "path", "start_streams",
                                       "config_files", "api_file", "enabled_groups"])


def parse_args(args) -> ParsedArgs:
    """
    Parse command line arguments:
    param: args: in form of --arg=value
    --path, optional, is the path of posix socket
    --port, optional the tcp port number
    --start-streams, optional True if to auto start all events of STREAM type
    --config-files, is a comma-separated list of hopeit apps config files relative or full paths
    --api-file, optional path to openapi.json file with at least openapi and info sections
    --enabled-groups, optional list of group label to be started
    Example::

        python web.py --port=8020 --path=/tmp/hopeit.01 --config-files=test.json

    Notes:
        --config-files argument is mandatory
        if --port and --path are not supplied the engine start on 8020 port by default

    """
    parser = argparse.ArgumentParser(description="hopeit.py engine")
    parser.add_argument('--host')
    parser.add_argument('--path')
    parser.add_argument('--port')
    parser.add_argument('--start-streams', action='store_true')
    parser.add_argument('--config-files')
    parser.add_argument('--api-file')
    parser.add_argument('--enabled-groups')

    parsed_args = parser.parse_args(args=args)
    port = int(parsed_args.port) if parsed_args.port else 8020 if parsed_args.path is None else None
    config_files = parsed_args.config_files.split(',')
    enabled_groups = parsed_args.enabled_groups.split(',') if parsed_args.enabled_groups else []

    return ParsedArgs(
        host=parsed_args.host,
        port=port,
        path=parsed_args.path,
        start_streams=bool(parsed_args.start_streams),
        config_files=config_files,
        api_file=parsed_args.api_file,
        enabled_groups=enabled_groups
    )


if __name__ == "__main__":
    sys_args = parse_args(sys.argv[1:])
    prepare_engine(
        config_files=sys_args.config_files,
        api_file=sys_args.api_file,
        start_streams=sys_args.start_streams,
        enabled_groups=sys_args.enabled_groups
    )
    serve(host=sys_args.host, path=sys_args.path, port=sys_args.port)



"""
API Definition helpers for user apps
"""
import inspect
from functools import partial
from typing import Optional, List, Type, Dict, Callable, Union, Tuple, Any, TypeVar

import re
import typing_inspect  # type: ignore

from hopeit.dataobjects import BinaryDownload
from hopeit.app.config import AppConfig, AppDescriptor, EventType
from hopeit.server.api import spec, app_route_name, APIError, BUILTIN_TYPES, datatype_schema
from hopeit.server.names import route_name

__all__ = ['api_from_config',
           'event_api',
           'app_base_route_name']

ArgType = TypeVar("ArgType", bound=Type)  # pylint: disable=invalid-name
ArgDef = Union[str, Tuple[str, ArgType], Tuple[str, ArgType, str]]
PayloadDef = Union[Type, Tuple[Type, str]]


def api_from_config(module, *, app_config: AppConfig, event_name: str, plugin: Optional[AppConfig]) -> dict:
    """
    Uses api definition exactly as it comes from --api-file specified at runtime.

    Usage in app event implementation file::

        __api__ = api_from_config

    Notice that no arguments need to be provided (just the function) as the function will be invoked
    during server initialization.
    """
    assert spec is not None
    plugin_app = None if plugin is None else plugin.app
    override_route_name = app_config.events[event_name].route
    route = app_route_name(app_config.app, event_name=event_name, plugin=plugin_app,
                           override_route_name=override_route_name)
    method = app_config.events[event_name].type.value.lower()
    event_spec = spec['paths'].get(route)
    if event_spec is None:
        raise APIError(f"Missing API Config for app={app_config.app.app_key()} event={event_name}")
    method_spec = event_spec.get(method)
    if method_spec is None:
        raise APIError(f"Missing method={method} for app={app_config.app.app_key()} event={event_name}")
    return event_spec[method]


def app_base_route_name(app: AppDescriptor, *, plugin: Optional[AppDescriptor] = None, prefix='api') -> str:
    """
    Returns base route name for paths in a given app
    """
    components = [
        prefix, app.name, app.version,
        *([plugin.name, plugin.version] if plugin else [])
    ]
    return route_name(*components)


def _arg_name(arg: ArgDef) -> str:
    if isinstance(arg, str):
        return arg
    return arg[0]


def _arg_type(arg: ArgDef) -> Tuple[str, Optional[str], bool]:
    if isinstance(arg, str):
        return (*BUILTIN_TYPES[str], True)
    datatype, required = arg[1], True
    origin = typing_inspect.get_origin(datatype)
    if origin is Union:
        type_args = typing_inspect.get_args(datatype)
        datatype = type_args[0]
        required = type_args[-1] is None
    return (*BUILTIN_TYPES.get(datatype, BUILTIN_TYPES[str]), required)


def _arg_description(arg: ArgDef) -> str:
    if isinstance(arg, str):
        return arg
    if len(arg) == 3:
        return arg[2]  # type: ignore
    return arg[0]


def _payload_schema(event_name: str, arg: PayloadDef) -> dict:
    datatype = arg[0] if isinstance(arg, tuple) else arg
    return datatype_schema(event_name, datatype)


def _payload_description(arg: PayloadDef) -> str:
    if isinstance(arg, tuple):
        return arg[1]
    if hasattr(arg, '__name__'):
        return arg.__name__
    return str(arg)


def _method_summary(module: str, summary: Optional[str] = None) -> str:
    if summary is not None:
        return summary
    doc_str = inspect.getdoc(module)
    if doc_str is not None:
        return doc_str.split("\n", maxsplit=1)[0]
    return ""


def _method_description(module: str, description: Optional[str] = None, summary: Optional[str] = None) -> str:
    if description is not None:
        return description
    doc_str = inspect.getdoc(module)
    if doc_str is not None and doc_str.count('\n') > 1:
        return re.sub(r"^\W+", "", doc_str.split("\n", 1)[1])
    return _method_summary(module, summary)


def _parse_args_schema(query_args: Optional[List[ArgDef]]):
    """
    Parse query args schema
    """
    parameters = []
    for query_arg in (query_args or []):
        arg_name = _arg_name(query_arg)
        arg_type, arg_format, arg_req = _arg_type(query_arg)
        arg_desc = _arg_description(query_arg)
        arg_spec: dict = {
            "name": arg_name,
            "in": "query",
            "required": arg_req,
            "description": arg_desc,
            "schema": {
                "type": arg_type
            }
        }
        if arg_format is not None:
            arg_spec["schema"]["format"] = arg_format
        parameters.append(arg_spec)
    return parameters


def _parse_fields_schema(fields: Optional[List[ArgDef]]):
    """
    Parse form fields schema
    """
    fields_schema = {"type": "object", "required": [], "properties": {}}
    encoding = {}
    for field in (fields or []):
        arg_name = _arg_name(field)
        arg_type = str if len(field) == 1 else field[1]
        arg_schema = datatype_schema(arg_name, arg_type)  # type: ignore
        fields_schema['required'].append(arg_name)  # type: ignore
        props = arg_schema.get('properties', {arg_name: arg_schema})
        fields_schema['properties'].update(props)  # type: ignore
        if props[arg_name].get('type') == 'string':
            encoding[arg_name] = {
                'contentType':
                'application/octect-stream' if props[arg_name].get('format') == 'binary' else 'text/plain'
            }
        else:
            encoding[arg_name] = {'contentType': 'application/json'}
        arg_desc = _arg_description(field)
        if arg_desc is not None:
            fields_schema['properties'][arg_name]['description'] = arg_desc  # type: ignore
    return fields_schema, encoding


def _event_api(
        summary: Optional[str],
        description: Optional[str],
        payload: Optional[Type],
        query_args: Optional[List[ArgDef]],
        fields: Optional[List[ArgDef]],
        responses: Optional[Dict[int, PayloadDef]],
        module, app_config: AppConfig, event_name: str, plugin: Optional[AppConfig]) -> dict:
    """
    Handler returned by event_api(...)
    """
    parameters = _parse_args_schema(query_args)

    method_spec: Dict[str, Any] = {
        "summary": _method_summary(module, summary),
        "description": _method_description(module, description, summary),
        "parameters": parameters
    }

    event_config = app_config.events[event_name]
    content_type = 'multipart/form-data' if event_config.type == EventType.MULTIPART else 'application/json'
    if payload is not None:
        method_spec['requestBody'] = {
            "description": _payload_description(payload),
            "required": True,
            "content": {
                content_type: {
                    "schema": _payload_schema(event_name, payload)
                }
            }
        }

    def _payload_content_type(arg: PayloadDef) -> Tuple[PayloadDef, str]:
        dt = arg[0] if isinstance(arg, tuple) else arg
        if inspect.isclass(dt) and issubclass(dt, BinaryDownload):
            return (BinaryDownload, arg[1]) if isinstance(arg, tuple) else BinaryDownload, dt.content_type
        return arg, "application/json"

    def _response_content(datatype) -> dict:
        dt, content_type = _payload_content_type(datatype)
        return {content_type: {
                "schema": _payload_schema(event_name, dt)}}

    if fields is not None:
        if payload is not None:
            raise APIError("Payload and fields cannot be specified at the same time.")

        method_schema, encoding = _parse_fields_schema(fields)
        method_spec['requestBody'] = {
            "required": True,
            "content": {
                content_type: {
                    "schema": method_schema,
                    "encoding": encoding
                }
            }
        }

    api_responses = {
        str(status): {
            "description": _payload_description(datatype),
            "content": _response_content(datatype)
        } for status, datatype in (responses or {}).items()}
    method_spec['responses'] = api_responses
    return method_spec


def event_api(*, summary: Optional[str] = None,
              description: Optional[str] = None,
              payload: Optional[PayloadDef] = None,
              query_args: Optional[List[ArgDef]] = None,
              fields: Optional[List[ArgDef]] = None,
              responses: Optional[Dict[int, PayloadDef]] = None) -> Callable[..., dict]:
    """
    Provides a convenient way to define Open API specification using Python types for a given app event
    implementation module.

    :param summary: An optional, string summary. If not provided will be taken from module docstring first line.
    :param description: An optional, string description. If not provided will be taken from module docstring.
    :param payload: Payload schema definition. Could be a single data type, or a tuple with a Type and a description.
    :param query_args: List of query arguments: each argument could be a single string with the arg name (in which case
        str type will be assumed), or a tuple of (str, type), where type if a valid datatype for query args (str, int,
        float, bool), or a tuple of (str, type, str) where last string is argument description.
    :param responses: a dictionary where key HTTP status code and value is the payload definition as describer in
        payload parameter.

    Examples:

    This will generate Open Schema with default types and description::

        __api__ = event_api(payload=CustomDataType, query_args['arg1'], responses={200: CustomResultDataType})

    Types and descriptions can be defined more precisely, and multiple response types can be specified::

        __api__ = event_api(
            payload=(CustomDataType, "A CustomDataType object to be used as an input"),
            query_args[('arg1', str, "Argument expected in query string",
            responses: {
                200: (CustomResultDataType, "Result in case operation is successful"),
                404: (NotFoundInformation, "Information return in case object is not found")
            }
        )
    """

    return partial(_event_api, summary, description, payload, query_args, fields, responses)



"""
hopeit.engine testing module

Provides utilities to test and write unit or integration tests for App Events:

    * **apps**: load config and execute app events for testing behaviour. Allows execution
      of events without starting a server.
    * **encryption**: provides data encryption for tests. Useful to test data apps.
"""
__all__ = ['apps',
           'encryption']



import setuptools

version = {}
with open("src/hopeit/server/version.py") as fp:
    exec(fp.read(), version)


def read_requirements_txt():
    with open("requirements.txt") as fb:
        libs = {}
        for line in fb.readlines():
            for op in (">=", "=="):
                try:
                    idx = line.index(op)
                    libs[line[0:idx]] = line[idx+2:]
                except ValueError:
                    pass
    return libs

def read_requirements_lock():
    with open("requirements.lock") as fb:
        libs = {}
        for line in fb.readlines():
            lv = line.split("==")
            if len(lv) >  1:
                libs[lv[0]] = lv[1].strip('\n')
    return libs

req_versions = read_requirements_txt()
locked_versions = read_requirements_lock()


def libversion(lib):
    lib_source = "requirements.txt"
    lib_version = req_versions.get(lib)
    if lib_version is None:
        lib_source = "requirements.lock"
        lib_version = locked_versions[lib.split('[')[0]]
    print(lib_source, f"{lib}>={lib_version}")
    return lib_version


setuptools.setup(
    name="hopeit.engine",
    version=version['ENGINE_VERSION'],
    description="Hopeit Engine: Microservices with Streams",
    license="Apache 2",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author="Leo Smerling and Pablo Canto",
    author_email="contact@hopeit.com.ar",
    url="https://github.com/hopeit-git/hopeit.engine",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: AsyncIO",
    ],    
    project_urls={
        "CI: GitHub Actions": "https://github.com/hopeit-git/hopeit.engine/actions?query=workflow",  # noqa
        "Docs: RTD": "https://hopeitengine.readthedocs.io/en/latest/",
        "GitHub: issues": "https://github.com/hopeit-git/hopeit.engine/issues",
        "GitHub: repo": "https://github.com/hopeit-git/hopeit.engine",
    },
    package_dir={
        "": "src"
    },
    packages=[
        "hopeit.app",
        "hopeit.cli",
        "hopeit.dataobjects",
        "hopeit.server",
        "hopeit.streams",
        "hopeit.testing",
        "hopeit.toolkit"
    ],
    include_package_data=True,
    package_data={
        "hopeit.app": ["py.typed"],
        "hopeit.cli": ["py.typed"],
        "hopeit.dataobjects": ["py.typed"],
        "hopeit.server": ["py.typed"],
        "hopeit.streams": ["py.typed"],
        "hopeit.testing": ["py.typed"],
        "hopeit.toolkit": ["py.typed"]
    },
    python_requires=">=3.7",
    install_requires=[ f"{lib}>={libversion(lib)}" for lib in [
        "lz4",
        "stringcase",
        "PyJWT[crypto]",
        "deepdiff",
        "typing-inspect",
        "multidict",
        "dataclasses-jsonschema[fast-validation]",
        "fastjsonschema"
    ]],
    extras_require={
        "web": [ f"{lib}>={libversion(lib)}" for lib in [
            "aiohttp",
            "aiohttp-cors",
            "aiohttp-swagger3"
        ]],
        "cli": [ f"{lib}>={libversion(lib)}" for lib in [
            "click"
        ]],
        "redis-streams": [
            f"hopeit.redis-streams=={version['ENGINE_VERSION']}"
        ],
        "redis-storage": [
            f"hopeit.redis-storage=={version['ENGINE_VERSION']}"
        ],
        "fs-storage": [
            f"hopeit.fs-storage=={version['ENGINE_VERSION']}"
        ],
        "config-manager": [
            f"hopeit.config-manager=={version['ENGINE_VERSION']}"
        ],
        "log-streamer": [
            f"hopeit.log-streamer=={version['ENGINE_VERSION']}"
        ],
        "apps-visualizer": [
            f"hopeit.apps-visualizer=={version['ENGINE_VERSION']}"
        ],
        "apps-client": [
            f"hopeit.apps-client=={version['ENGINE_VERSION']}"
        ]
    },
    entry_points={
        "console_scripts": [
            "hopeit_server = hopeit.cli.server:server",
            "hopeit_openapi = hopeit.cli.openapi:openapi"
        ]
    }
)



"""
Context information and handling
"""
from typing import AsyncGenerator, AsyncIterator, Callable, Generic, TypeVar, \
    Dict, Optional, Any, Tuple, Union, List
from abc import ABC
from pathlib import Path
from datetime import datetime, timezone
import re

from aiohttp import web
from multidict import MultiDict, CIMultiDict, CIMultiDictProxy, istr
from stringcase import titlecase  # type: ignore

from hopeit.app.config import AppConfig, AppDescriptor, EventDescriptor, Env, EventSettings, EventType


__all__ = ['EventContext',
           'PostprocessHook',
           'PreprocessHook',
           'NoopMultiparReader']


class EventContext:
    """
    EventContext information, shared between engine components
    when an event is executed. Instances of this class are
    created automatically by the engine on every request.

    Instantiated objects contain static information during app lifecycle:
    :param app_key: app key string of the engine app running
    :param app: AppDescriptor of the engine app running
    :param env: dict of env configuration and dynamic information per event execution:
    :param event_name: str
    :param event_info: EventDescriptor of the event executed
    :param track_ids: dict of keys and id values to be tracked
    """
    def __init__(self, *,
                 app_config: AppConfig,
                 plugin_config: AppConfig,
                 event_name: str,
                 settings: EventSettings,
                 track_ids: Dict[str, str],
                 auth_info: Dict[str, Any]):
        self.app_key: str = app_config.app_key()
        self.plugin_key: str = plugin_config.app_key()
        self.app: AppDescriptor = app_config.app
        self.env: Env = {**plugin_config.env, **app_config.env}
        self.event_name = event_name
        self.settings = settings
        base_event = event_name.split('$')[0]
        self.event_info: EventDescriptor = plugin_config.events[base_event]
        self.creation_ts: datetime = datetime.now(tz=timezone.utc)
        self.auth_info = auth_info
        track_fields = [
            'track.operation_id',
            'track.client_app_key',
            'track.client_event_name',
            *app_config.engine.track_headers,
            *(settings.logging.stream_fields
              if self.event_info.type == EventType.STREAM
              else [])
        ]
        self.track_ids = {
            k: track_ids[k] for k in track_fields if k in track_ids
        }
        self.track_ids['event.app'] = self.app_key
        if self.plugin_key != self.app_key:
            self.track_ids['event.plugin'] = self.plugin_key


class PostprocessStreamResponseHook():
    """
    Post process stream response hook

    Useful to stream content and avoid memory overhead.
    """
    def __init__(self, content_disposition: str, content_type: str, content_length: int):
        self.headers = MultiDict({
            "Content-Disposition": content_disposition,
            "Content-Type": content_type,
        })
        self.resp = web.StreamResponse(headers=self.headers)
        self.resp.content_type = content_type
        self.resp.content_length = content_length

    async def prepare(self, request: web.Request):
        await self.resp.prepare(request)

    async def write(self, data: bytes):
        await self.resp.write(data)


class TestingResp:
    def __init__(self):
        self.headers = MultiDict()
        self.data: bytes = b''


class PostprocessTestingStreamResponseHook(PostprocessStreamResponseHook):
    """
    Post process stream response hook

    Useful to stream content and avoid memory overhead.
    """
    def __init__(self, content_disposition: str, content_type: str, content_length: int):
        super().__init__(content_disposition, content_type, content_length)
        self.resp: TestingResp = TestingResp()  # type: ignore
        self.headers = MultiDict({
            "Content-Disposition": content_disposition,
            "Content-Type": content_type,
            "Content-Length": str(content_length),
        })

    async def prepare(self, request):
        self.resp.data = b''

    async def write(self, data: bytes):
        self.resp.data += data


class PostprocessHook():
    """
    Post process hook that keeps additional changes to add to response on
    `__postprocess__(...)` event methods.

    Useful to set cookies, change status and set additional headers in web responses.
    """
    def __init__(self, request: Optional[web.BaseRequest] = None):
        self.headers: Dict[str, str] = {}
        self.cookies: Dict[str, Tuple[str, tuple, dict]] = {}
        self.del_cookies: List[Tuple[str, tuple, dict]] = []
        self.status: Optional[int] = None
        self.file_response: Optional[Union[str, Path]] = None
        self.stream_response: Optional[PostprocessStreamResponseHook] = None
        self.content_type: str = "application/json"
        self.request = request

    async def prepare_stream_response(
        self, context: EventContext, content_disposition: str, content_type: str, content_length: int
    ):
        """
        Prepare stream response allow to send file like objects as stream response.
        """
        if self.request:
            self.stream_response = PostprocessStreamResponseHook(content_disposition, content_type, content_length)
        else:
            self.stream_response = PostprocessTestingStreamResponseHook(content_disposition,
                                                                        content_type, content_length)
        self.headers.update(
            self.stream_response.headers
        )
        self.stream_response.resp.headers.update({
            **self.headers,
            **{f"X-{re.sub(' ', '-', titlecase(k))}": v for k, v in context.track_ids.items()}
        })
        self.content_type = self.stream_response.headers["Content-Type"]

        await self.stream_response.prepare(self.request)  # type: ignore
        return self.stream_response

    def set_header(self, name: str, value: Any):
        self.headers[name] = value

    def set_content_type(self, content_type: str):
        self.content_type = content_type

    def set_cookie(self, name: str, value: str, *args, **kwargs):
        self.cookies[name] = (value, args, kwargs)

    def del_cookie(self, name: str, *args, **kwargs):
        self.del_cookies.append((name, args, kwargs))

    def set_status(self, status: int):
        self.status = status

    def set_file_response(self, path: Union[str, Path]):
        self.file_response = path


class BodyPartReaderProtocol(ABC):
    """
    Required functionallity for BodyPartReader implementation
    to be used in PreprocessHook
    """
    async def read_chunk(self, size: Optional[int] = None) -> bytes:
        ...

    async def text(self, *, encoding: Optional[str] = None) -> str:
        ...

    async def json(self, *, encoding: Optional[str] = None) -> Optional[Dict[str, Any]]:
        ...

    @property
    def name(self) -> Optional[str]:
        ...

    @property
    def filename(self) -> Optional[str]:
        ...

    @property
    def headers(self) -> CIMultiDictProxy[str]:
        ...


class MultipartReaderProtocol(ABC):
    """
    Required functionality for Mutipartreader implementation
    to be used in PreprocessHook
    """

    def __aiter__(
        self,
    ) -> AsyncIterator["BodyPartReaderProtocol"]:
        ...

    async def __anext__(
        self,
    ) -> Optional[Union["MultipartReaderProtocol", BodyPartReaderProtocol]]:
        ...


class NoopMultiparReader(MultipartReaderProtocol):
    pass


_MultipartReader = TypeVar('_MultipartReader', bound=MultipartReaderProtocol)
_BodyPartReader = TypeVar('_BodyPartReader', bound=BodyPartReaderProtocol)


class PreprocessFileHook(Generic[_BodyPartReader]):
    """
    Hook to read files from multipart requests
    """
    def __init__(self, *, name: str, file_name: str, data: _BodyPartReader):
        self.name = name
        self.file_name = file_name
        self.data = data
        self.size = 0

    async def _read_chunk(self, *, chunk_size: int) -> bytes:
        chunk = await self.data.read_chunk(size=chunk_size)
        self.size += len(chunk)
        return chunk

    async def read(self, chunk_size: int = -1) -> bytes:
        """
        File like object read function

        :param chunk_size: Size in bytes of each chunk.
            If chunk_size > 0 read will return a part of upcoming file of chunk_size size
            If chunk_size = -1 read will return the full upcoming file
            If chunk_size = 0 read will return empty bytes
        :return bytes
        """
        if chunk_size > 0:
            return await self._read_chunk(chunk_size=chunk_size)
        if chunk_size == -1:
            content = b""
            async for chunk in self.read_chunks(chunk_size=chunk_size):
                content += chunk
            return content
        return b""

    async def read_chunks(self, *, chunk_size: int) -> AsyncGenerator[bytes, None]:
        """
        Returns from multipart requests a file in chunks as generator:

        :param chunk_size: Size in bytes of each chunk
        """
        chunk = await self._read_chunk(chunk_size=chunk_size)
        while chunk:
            yield chunk
            chunk = await self._read_chunk(chunk_size=chunk_size)


class PreprocessHeaders:
    """
    Wrapper to receive request headers in `__preprocess__` functions
    """
    def __init__(self, request_headers: CIMultiDictProxy[str]) -> None:
        self._headers = request_headers

    def __getitem__(self, key: str) -> Any:
        return self._headers[key]

    def get(self, key: str) -> Any:
        return self._headers.get(key)

    def __repr__(self):
        return self._headers.__repr__()

    @classmethod
    def from_dict(cls, data: Dict[str, str]):
        return cls(CIMultiDictProxy(CIMultiDict(data)))


class PreprocessHook(Generic[_MultipartReader]):
    """
    Preprocess hook that handles information available in the request to be accessed
    from `__preprocess__(...)` event method when defined.
    """
    def __init__(self, *, headers: CIMultiDictProxy[str],
                 payload_raw: Optional[bytes] = None,
                 multipart_reader: Optional[_MultipartReader] = None,
                 file_hook_factory: Callable = PreprocessFileHook):
        self.headers = PreprocessHeaders(headers)
        self.payload_raw = payload_raw
        self._multipart_reader = multipart_reader
        self._args: Dict[str, Any] = {}
        self._iterated = False
        self.status: Optional[int] = None
        self.file_hook_factory = file_hook_factory

    def set_status(self, status: int):
        self.status = status

    async def parsed_args(self):
        if not self._iterated:
            async for _ in self.files():
                pass
        return self._args

    async def files(self) -> AsyncGenerator[Any, None]:
        """
        Iterator over attached files in multipart uploads
        """
        assert not self._iterated, "Request fields already extracted"
        self._iterated = True
        if self._multipart_reader is not None:
            async for field in self._multipart_reader:
                if field.name is not None:
                    if field.filename:
                        self._args[field.name] = field.filename
                        yield self.file_hook_factory(name=field.name, file_name=field.filename, data=field)
                    elif field.headers.get(istr("Content-Type")) == 'application/json':
                        self._args[field.name] = await field.json()
                    else:
                        self._args[field.name] = await field.text()



"""
auth module, provides functionallity to create, validate and decode access and refresh tokens

call init() to load or create RSA keys before using token manipulation methods
"""
from typing import Dict, Optional, Set

import pathlib
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization import load_pem_public_key

import jwt
from jwt import InvalidSignatureError, ExpiredSignatureError, DecodeError

from hopeit.app.context import EventContext
from hopeit.server.config import AuthConfig, AuthType
from hopeit.server.logger import engine_extra_logger

__all__ = ['init',
           'new_token',
           'decode_token',
           'validate_token',
           'validate_auth_method',
           'AuthType']

logger, extra = engine_extra_logger()

auth_config: Optional[AuthConfig] = None
private_keys: Dict[str, RSAPrivateKey] = {}
public_keys: Dict[str, RSAPublicKey] = {}
rejected_apps: Set[str] = set()


def init(app_key: str, app_auth_config: AuthConfig):
    """
    loads or creates RSA private/public keys that will be used to create, validate and decode tokens

    :param auth_config: AuthConfig, authorization/authentication server configuration
    """
    global public_keys, private_keys, auth_config
    logger.info(__name__, "Initializing auth module...", extra=extra(app_key=app_key))
    auth_config = app_auth_config
    secrets_path = pathlib.Path(auth_config.secrets_location)
    if auth_config.enabled:
        private_key_path = secrets_path / '.private' / f'{app_key}.pem'
        public_key_path = secrets_path / 'public' / f'{app_key}_pub.pem'
        passphrase = auth_config.auth_passphrase.encode()
        try:
            with open(private_key_path, 'rb') as f:
                private_keys[app_key] = load_pem_private_key(f.read(), passphrase, default_backend())  # type: ignore
            with open(public_key_path, 'rb') as f:
                public_keys[app_key] = load_pem_public_key(f.read(), default_backend())  # type: ignore
            logger.info(__name__, f"Found keys in {auth_config.secrets_location}", extra=extra(app_key=app_key))
        except FileNotFoundError as e:
            if auth_config.create_keys:
                logger.info(__name__, "Keys not found. Creating new key pair...", extra=extra(app_key=app_key))
                os.makedirs(secrets_path / '.private', exist_ok=True)
                os.makedirs(secrets_path / 'public', exist_ok=True)
                private_key = rsa.generate_private_key(
                    public_exponent=65537,
                    key_size=2048,
                    backend=default_backend()
                )
                public_key = private_key.public_key()
                with open(private_key_path, "wb") as f:
                    f.write(private_key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.BestAvailableEncryption(passphrase),
                    ))
                with open(public_key_path, "wb") as f:
                    f.write(public_key.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
                    ))
                public_keys[app_key] = public_key
                private_keys[app_key] = private_key
                logger.debug(__name__, "Saved generated key pair.", extra=extra(app_key=app_key))
            else:
                raise e
    else:
        public_keys = {}
        private_keys = {}
        logger.warning(__name__, 'Auth module not configured in server config. Auth disabled.',
                       extra=extra(app_key=app_key))


class PublicKeyNotFoundError(Exception):
    pass


class PrivateKeyNotFoundError(Exception):
    pass


def app_public_key(app_key: str) -> RSAPublicKey:
    """
    Returns current registerd public key for a given application.
    Attempts to load from disk in case is not cached.
    If previous attemps to load key failed, it will not attempt to load again,
    so an app restart might be necessary in case configuration is fixed.
    """
    def _load_app_public_key():
        if app_key in rejected_apps:
            raise PublicKeyNotFoundError(app_key)
        try:
            logger.info(__name__, "Loading public app keys", extra=extra(app_key=app_key))
            public_key_path = pathlib.Path(auth_config.secrets_location) / 'public' / f'{app_key}_pub.pem'
            with open(public_key_path, 'rb') as f:
                public_keys[app_key] = load_pem_public_key(
                    f.read(), default_backend()
                )
            return public_keys[app_key]
        except FileNotFoundError as e:
            logger.error(__name__, "Cannot find public key for app", extra=extra(app_key=app_key))
            rejected_apps.add(app_key)
            raise PublicKeyNotFoundError(app_key) from e

    key = public_keys.get(app_key)
    if key is None:
        return _load_app_public_key()
    return key


def app_private_key(app_key: str) -> RSAPrivateKey:
    key = private_keys.get(app_key)
    if key is None:
        raise PrivateKeyNotFoundError(app_key)
    return key


def new_token(app_key: str, payload: dict) -> str:
    private_key = app_private_key(app_key)
    return jwt.encode({**payload, "app": app_key}, private_key, algorithm='RS256')  # type: ignore


def decode_token(token: str) -> dict:
    info = jwt.decode(token, algorithms=['RS256'], options={"verify_signature": False})
    app_key = info['app']
    public_key = app_public_key(app_key)
    return jwt.decode(token, public_key, algorithms=['RS256'])  # type: ignore


def validate_token(token: str, context: EventContext) -> Optional[dict]:
    try:
        return decode_token(token)
    except (InvalidSignatureError,
            ExpiredSignatureError,
            DecodeError,
            PublicKeyNotFoundError) as e:
        logger.warning(context, e)
        return None


def validate_auth_none(data: str, context: EventContext):
    context.auth_info['auth_type'] = AuthType.UNSECURED
    context.auth_info['allowed'] = True


def validate_auth_basic(data: str, context: EventContext):
    if data:
        context.auth_info['auth_type'] = AuthType.BASIC
        context.auth_info['allowed'] = True
        context.auth_info['payload'] = data


def validate_auth_bearer(data: str, context: EventContext):
    auth_info = validate_token(data, context)
    if auth_info is not None:
        context.auth_info['auth_type'] = AuthType.BEARER
        context.auth_info['allowed'] = True
        context.auth_info['payload'] = auth_info


def validate_auth_refresh(data: str, context: EventContext):
    auth_info = validate_token(data, context)
    if auth_info is not None:
        context.auth_info['auth_type'] = AuthType.REFRESH
        context.auth_info['allowed'] = True
        context.auth_info['payload'] = auth_info


AUTH_VALIDATION_METHODS = {
    AuthType.UNSECURED: validate_auth_none,
    AuthType.BASIC: validate_auth_basic,
    AuthType.BEARER: validate_auth_bearer,
    AuthType.REFRESH: validate_auth_refresh
}


def validate_auth_method(auth_type: AuthType, data: str, context: EventContext):
    return AUTH_VALIDATION_METHODS[auth_type](data, context)



"""
Handling sequencing and execution of steps from events
"""
import asyncio
from datetime import datetime, timezone
from functools import partial
from types import ModuleType
from typing import Type, Dict, Any, Optional, Callable, Tuple, AsyncGenerator, List, Union
import inspect

from hopeit.app.config import AppConfig, AppDescriptor, EventDescriptor, EventType, \
    ReadStreamDescriptor, StreamQueueStrategy, WriteStreamDescriptor
from hopeit.app.context import EventContext
from hopeit.dataobjects import DataObject, EventPayload, copy_payload, EventPayloadType
from hopeit.server.imports import find_event_handler
from hopeit.server.logger import engine_logger, extra_logger
from hopeit.server.names import auto_path
from hopeit.server.collector import Collector, AsyncCollector, CollectorStepType

__all__ = ['extract_module_steps',
           'effective_steps',
           'event_and_step',
           'extract_event_stages',
           'extract_preprocess_handler',
           'extract_postprocess_handler',
           'extract_input_type',
           'execute_steps',
           'invoke_single_step',
           'find_datatype_handler',
           'StepInfo',
           'split_event_stages',
           'SHUFFLE',
           'CollectorStepsDescriptor']


SHUFFLE = '__shuffle__'
MAX_STEPS = 1000  # To protect for infinite loops up to 1000 sequential steps can be executed

StepInfo = Tuple[Callable, Type, Type, bool]
StepExecutionList = List[Tuple[int, str, StepInfo]]

logger = engine_logger()
extra = extra_logger()


def extract_module_steps(impl: ModuleType) -> List[Tuple[str, Optional[StepInfo]]]:
    assert hasattr(impl, '__steps__'), f"Missing `__steps__` definition in module={impl.__name__}"
    return [
        (step_name, None if step_name == SHUFFLE else _signature(impl, step_name))
        for step_name in getattr(impl, '__steps__')
    ]


def effective_steps(event_name: str,
                    module_steps: List[Tuple[str, Optional[StepInfo]]]) -> StepExecutionList:
    """
    Return list of steps given the possibility that event steps are splitted in stages
    using SHUFFLE keyword. In that case event name will contain `event_name.stage_step` format.
    In case event_name does not contain '.' and stage_step, the list of steps from start
    up to a SHUFFLE if found is returned. In case event_name hast '.' and stage_step,
    steps starting in stage_step up to the next SHUFFLE step if present is are returned.
    """
    i = 0
    steps: StepExecutionList = []
    _, from_step = event_and_step(event_name)
    found = (from_step is None)
    for step_name, step_info in module_steps:
        if found:
            if step_name == SHUFFLE:
                break
            assert step_info
            steps.append((i, step_name, step_info))
        elif step_name == from_step:
            i = 0
            found = True
            assert step_info
            steps.append((i, step_name, step_info))
        i += 1
    return steps


def extract_event_stages(impl: ModuleType) -> List[str]:
    """
    Extract a list of step names consisting of the first step defined in __steps__ plus
    the subsequent step name after a SHUFFLE step if any in a given module
    """
    assert hasattr(impl, '__steps__'), f"Missing `__steps__` definition in module={impl.__name__}"
    stages = []
    steps = getattr(impl, '__steps__')
    if len(steps) == 1:
        return steps
    stage = steps[0] if isinstance(steps[0], str) else steps[0].__name__
    for step_name in steps[1:]:
        if step_name == SHUFFLE:
            assert stage, f"Invalid location of `SHUFFLE` step in module={impl.__name__}"
            stages.append(stage)
            stage = None
        elif stage is None:
            stage = step_name
    if stage:
        stages.append(stage)
    return stages


def extract_postprocess_handler(impl: ModuleType) -> Optional[StepInfo]:
    if hasattr(impl, '__postprocess__'):
        return _signature(impl, '__postprocess__')
    return None


def extract_preprocess_handler(impl: ModuleType) -> Optional[StepInfo]:
    if hasattr(impl, '__preprocess__'):
        return _signature(impl, '__preprocess__')
    return None


def extract_input_type(impl: ModuleType, from_step: Optional[str] = None) -> Type:
    assert hasattr(impl, '__steps__'), f"Missing `__steps__` definition in module={impl.__name__}"
    if from_step is None and hasattr(impl, '__preprocess__'):
        _, input_type, _, _ = _signature(impl, '__preprocess__')
        return input_type
    for step_name in getattr(impl, '__steps__'):
        if from_step is None or step_name == from_step:
            _, input_type, _, _ = _signature(impl, step_name)
            return input_type
    raise NotImplementedError(f"Cannot find payload datatype for module={impl.__name__} step={from_step}")


def event_and_step(event_name: str) -> Tuple[str, Optional[str]]:
    comps = event_name.split('$')
    if len(comps) > 1:
        return (
            comps[0],
            comps[1] if comps[1][0] != "_" else None  # Ignore special suffixes after '$' (i.e. `$__service__`)
        )
    return event_name, None


async def _execute_steps_recursion(payload: Optional[EventPayload],
                                   context: EventContext,
                                   steps: StepExecutionList,
                                   step_index: int,
                                   func: Optional[Callable],
                                   is_spawn: bool,
                                   step_delay: float,
                                   query_args: Dict[str, Any]) -> AsyncGenerator[Optional[EventPayload], None]:
    """
    Steps execution handler that allows processing Spawn events results using recursion,
    and sequential execution of events using iteration
    """
    if is_spawn:
        async for invoke_result in _invoke_spawn_step(
                copy_payload(payload),
                func,  # type: ignore
                context,
                **query_args):
            if step_delay:
                await asyncio.sleep(step_delay)
            i, f, it = _find_next_step(invoke_result, steps, from_index=step_index + 1)
            if i == -1:
                yield copy_payload(invoke_result)
            else:
                # Recursive call for each received element
                async for recursion_result in _execute_steps_recursion(
                    invoke_result, context, steps, i, f, it, step_delay, {}
                ):
                    yield recursion_result

    else:
        i, f, q, it, invoke_result = step_index, func, query_args, is_spawn, payload
        while 0 <= i < MAX_STEPS:
            # Recursive call if result is iterable (Spawn)
            if it:
                async for recursion_result in _execute_steps_recursion(
                    invoke_result, context, steps, i, f, it, step_delay, {}
                ):
                    yield recursion_result
                break

            # Single step invokation
            invoke_result = await _invoke_step(
                copy_payload(invoke_result),
                f,  # type: ignore
                context,
                **q
            )
            q = {}
            if step_delay:
                await asyncio.sleep(step_delay)
            i, f, it = _find_next_step(invoke_result, steps, from_index=i + 1)

        if i == -1:
            # Yields result if all steps were exhausted
            yield copy_payload(invoke_result)
        if i >= MAX_STEPS:
            raise RuntimeError(f"Maximun number of steps to execute exceeded (MAX_STEPS={MAX_STEPS}).")


async def execute_steps(steps: StepExecutionList,
                        *, context: EventContext,
                        payload: Optional[EventPayload],
                        **kwargs) -> AsyncGenerator[Optional[EventPayload], None]:
    """
    Invoke steps from a event.
    It will try to find next step in configuration order that matches input type of the payload,
    and will updated the payload and invoke next valid step.
    """
    start_ts = datetime.now(tz=timezone.utc)
    step_delay = context.settings.stream.step_delay / 1000.0
    throttle_ms = context.settings.stream.throttle_ms
    i, func, is_spawn = _find_next_step(payload, steps, from_index=0)
    if i >= 0:
        async for result in _execute_steps_recursion(
            payload, context, steps, i, func, is_spawn, step_delay, kwargs
        ):
            await _throttle(context, throttle_ms, start_ts)
            yield result


async def _invoke_step(payload: Optional[EventPayload],
                       func: Callable,
                       context: EventContext,
                       **kwargs) -> Optional[EventPayload]:
    """
    Invokes step handler method
    """
    func_res = func(payload, context, **kwargs)
    if inspect.iscoroutine(func_res):
        return await func_res
    return func_res


async def _invoke_spawn_step(payload: Optional[EventPayload],
                             func: Callable,
                             context: EventContext,
                             **kwargs) -> AsyncGenerator[Optional[EventPayload], None]:
    """
    Invokes step handler method that returns Spawn/AsyncGenerator type
    """
    async for res in func(payload, context, **kwargs):
        yield res


async def invoke_single_step(func: Callable, *,
                             payload: Optional[EventPayload],
                             context: EventContext,
                             **kwargs) -> Optional[EventPayload]:
    return await _invoke_step(payload=payload, func=func, context=context, **kwargs)


def _find_next_step(
    payload: Optional[EventPayload],
    steps: StepExecutionList,
    from_index: int
) -> Tuple[int, Optional[Callable], bool]:
    """
    Finds next step to exectute in pending_steps list, base on the payload data type
    """
    # Returns in case no more steps are available
    if from_index >= len(steps):
        return -1, None, False
    # Try to match specific payload datatypes incluiding None
    for i, _, step_info in steps[from_index:]:
        func, input_type, _, is_iterable = step_info
        if input_type is None and payload is None:
            return i, func, is_iterable
        if (
            input_type is not None
            and input_type is not DataObject
            and isinstance(payload, input_type)
        ):
            return i, func, is_iterable
    # Return if there was no match for None payload
    if payload is None:
        return -1, None, False
    # Finally try to match generic `payload: DataObject` argument in step
    for i, _, step_info in steps[from_index:]:
        func, input_type, _, is_iterable = step_info
        if input_type is DataObject and payload is not None:
            return i, func, is_iterable
    # No step input datatype matches current payload
    return -1, None, False


async def _throttle(context: EventContext, throttle_ms: int, start_ts: datetime):
    """
    Performs an async sleep in order to achieve duration specified by throttle configuration
    """
    delay = 0.0
    if throttle_ms:
        elapsed_td = datetime.now(tz=timezone.utc) - start_ts
        elapsed = 1000.0 * elapsed_td.total_seconds()
        delay = max(0.0, 1.0 * throttle_ms - elapsed)
    if delay > 0.0:
        logger.debug(context, "Throttling...",
                     extra=extra(prefix='throttle.', delay=delay, target=throttle_ms))
        await asyncio.sleep(delay / 1000.0)


def split_event_stages(app: AppDescriptor,
                       event_name: str,
                       event_info: EventDescriptor,
                       impl: ModuleType) -> Dict[str, EventDescriptor]:
    """
    Splits an event whose steps contain SHUFFLE step, in an initial event with same name as event_name
    plus sub_events with names `event_name.step_name' for each step after a SHUFFLE.
    Creates intermediate auto named write_stream, read_stream to communicate data between event and sub_events,
    clones event configuration from main event to sub_events, and setup write_stream property for final event
    to be the one specified in configuration.
    """
    event_stages = extract_event_stages(impl)
    if len(event_stages) == 1:
        return {event_name: event_info}

    effective_events: Dict[str, EventDescriptor] = {}
    event_type = event_info.type
    read_stream = event_info.read_stream
    queues = ["AUTO"] if read_stream is None else read_stream.queues
    sub_event_name: Optional[str] = event_name
    sub_event_info = event_info
    intermediate_stream = None
    for stage in event_stages:
        if sub_event_name is None:
            sub_event_name = f"{event_name}${stage}"
        if read_stream is None and intermediate_stream is not None:
            read_stream = ReadStreamDescriptor(
                name=intermediate_stream,
                consumer_group=auto_path(app.name, app.version, *event_name.split('.'), stage),
                queues=queues
            )
        intermediate_stream = auto_path(app.name, app.version, *event_name.split('.'), stage)
        sub_event_info = EventDescriptor(
            type=event_type,
            read_stream=read_stream,
            connections=event_info.connections,
            impl=event_info.impl,
            write_stream=WriteStreamDescriptor(
                name=intermediate_stream,
                queue_strategy=StreamQueueStrategy.PROPAGATE
            )
        )
        effective_events[sub_event_name] = sub_event_info
        event_type = EventType.STREAM
        sub_event_name = None
        read_stream = None
    # Set last stage write_stream to original event write_stream
    sub_event_info.write_stream = event_info.write_stream
    return effective_events


def find_datatype_handler(*, app_config: AppConfig, event_name: str, event_info: EventDescriptor):
    base_event, from_step = event_and_step(event_name)
    impl = find_event_handler(app_config=app_config, event_name=base_event, event_info=event_info)
    datatype = extract_input_type(impl, from_step=from_step)
    return datatype


class CollectorStepsDescriptor:
    """
    Specification for input payload type and list of sub-steps when using a collector as a compound steps
    in __steps__ definition of a method. This class should be instantiated using `collector` method
    from hopeit.app.events module.

    Example::

        from hopeit.app.events import collector_step

        __steps__ == [collector_step(payload=InputType).steps('step1', 'step2'), 'step_outside_collector']

    will generate a steps definition of two steps: First a collector, which receives an InputType object
    and run step1 and step2 functions concurrently. Step result will be a Collector to be used by
    `step_outside_collector`.
    """
    def __init__(self, input_type: Type[EventPayloadType]):
        self.input_type = input_type
        self.step_names: List[str] = []
        self.steps: Optional[List[Tuple[str, CollectorStepType]]] = None
        self.__name__ = f'collector@{id(self)}'

    def gather(self, *steps: str):
        self.step_names.extend(steps)
        self.__name__ = f'collector@{self.step_names[0]}'
        return self

    def __repr__(self) -> str:
        return f"collector(payload: {self.input_type.__name__}),\n  [" + \
               ',\n   '.join(f"'{name}'" for name in self.step_names) + ']'

    def name(self) -> str:
        return self.__name__

    def setup_step_impl(self, module: ModuleType) -> Callable[[EventPayload, EventContext], EventPayload]:
        if self.steps is None:
            self.steps = []
            for step_name in self.step_names:
                step_impl, payload_type, _, _ = _signature(module, step_name)
                assert step_impl is not None, "Collector can only contain function definitions."
                assert payload_type is AsyncCollector, f"step={step_name} first arg must be `Collector`"
                self.steps.append((step_name, step_impl))
        return partial(_run_collector, self)


async def _run_collector(step_group: CollectorStepsDescriptor, payload: EventPayload, context: EventContext):
    return await AsyncCollector().input(payload).steps(*step_group.steps).run(context)


def _signature(impl: ModuleType, step_name: Union[str, CollectorStepsDescriptor]) -> Tuple[Callable, Type, Type, bool]:
    """
    Computes signature (`CollectorStepsDescriptor`) from a given step (def) in an event module. Results
    are used internally by the engine to compute input/output datatypes of events and find proper steps
    to handle payloads and results.
    """
    if isinstance(step_name, CollectorStepsDescriptor):
        func = step_name.setup_step_impl(impl)
        annotation = step_name.input_type
        return_annotation = Collector
    else:
        func = getattr(impl, step_name)
        signature = inspect.signature(func)
        payload_arg = next(iter(signature.parameters.values()))
        annotation = payload_arg.annotation if payload_arg.annotation is not inspect.Signature.empty else Any
        return_annotation = signature.return_annotation if signature.return_annotation else Any
    return func, annotation, return_annotation, _is_iterable(return_annotation)


def _is_iterable(annotation: Type) -> bool:
    try:
        return annotation._name == "AsyncGenerator"  # pylint: disable=protected-access
    except AttributeError:
        return False



"""
Metrics module
"""
from datetime import datetime, timezone
from typing import Dict, Union, Optional

from hopeit.server.logger import format_extra_values
from hopeit.app.context import EventContext


__all__ = ['metrics',
           'stream_metrics',
           'StreamStats']


def metrics(context: EventContext):
    """
    Return event calculated metrics using EventContext as a dictionary that can be used
    in logging

    :param context: EventContext
    return: dictionary than can be passed to logging extra= parameter
    """
    return {'extra': format_extra_values(
        _calc_event_metrics(context),
        prefix='metrics.'
    )}


def stream_metrics(context: EventContext):
    """
    Return stream event calculated metrics using EventContext as a dictionary that can be used
    in logging

    :param context: EventContext
    :return: dictionary than can be passed to logging extra= parameter
    """
    return {'extra': format_extra_values(
        _calc_stream_metrics(context),
        prefix='metrics.'
    )}


def _calc_event_metrics(context: EventContext):
    duration = 1000.0 * (
        datetime.now(tz=timezone.utc) - context.creation_ts
    ).total_seconds()
    return {
        'duration': f"{duration:.3f}"
    }


def _calc_stream_metrics(context: EventContext):
    """
    Calculates stream events standard metrics:

    stream_age: difference between read_ts and submit_ts indicating time that event
        spent unconsumed
    event_elapsed: difference between now and event request_ts indicating time elapsed
        since event request was first created even before published to a stream
    :param: context, EventContext
    :return: dictionary than can be passed to logging extra= parameter
    """
    result = {}
    submit_ts = context.track_ids.get("stream.submit_ts")
    read_ts = context.track_ids.get("stream.read_ts")
    if submit_ts and read_ts:
        submit_dt = datetime.fromisoformat(submit_ts)
        read_dt = datetime.fromisoformat(read_ts)
        age = 1000.0 * (read_dt - submit_dt).total_seconds()
        result['stream_age'] = f"{age:.3f}"
    request_ts = context.track_ids.get("track.request_ts")
    if request_ts:
        request_dt = datetime.fromisoformat(request_ts)
        elapsed = 1000.0 * (
            datetime.now(tz=timezone.utc) - request_dt
        ).total_seconds()
        result['request_elapsed'] = f"{elapsed:.3f}"
    return result


class StreamStats:
    """
    Helper class to keep stream consuming stats
    """
    def __init__(self):
        self.run_ts: datetime = datetime.now(tz=timezone.utc)
        self.start_ts: Optional[datetime] = None
        self.from_ts: Optional[datetime] = None
        self.event_count: int = 0
        self.error_count: int = 0
        self.total_event_count = 0
        self.total_error_count: int = 0

    def ensure_start(self):
        if self.start_ts is None:
            self.start_ts = datetime.now(tz=timezone.utc)
            self.from_ts = datetime.now(tz=timezone.utc)
            self.event_count: int = 0
            self.error_count: int = 0
            self.total_event_count = 0
            self.total_error_count: int = 0
        return self

    def reset_batch(self, now: datetime):
        self.from_ts = now
        self.event_count = 0
        self.error_count = 0

    def inc(self, error: bool = False):
        self.event_count += 1
        self.total_event_count += 1
        if error:
            self.error_count += 1
            self.total_error_count += 1

    def calc(self) -> Dict[str, Union[int, float]]:
        """
        calculate stream stats to be logged
        :return: dict, with stream stats to be used as extra info for logging
        """
        assert self.start_ts is not None and self.from_ts is not None, \
            "StreamStats not initialized. Call `ensure_start()`"
        now = datetime.now(tz=timezone.utc)
        total_elapsed_td = now - self.start_ts
        total_elapsed = 1000.0 * total_elapsed_td.total_seconds()
        avg_rate = (1000.0 * self.total_event_count / total_elapsed) if total_elapsed else 0.0
        avg_rate_success = (1000.0 * (self.total_event_count - self.total_error_count) / total_elapsed) \
            if total_elapsed else 0.0
        avg_duration = (1.0 * total_elapsed / self.total_event_count) if self.total_event_count else 0.0
        partial_elapsed_td = now - self.from_ts
        partial_elapsed = 1000.0 * partial_elapsed_td.total_seconds()
        rate = (1000.0 * self.event_count / partial_elapsed) if partial_elapsed else 0.0
        rate_success = (1000.0 * (self.event_count - self.error_count) / partial_elapsed) \
            if partial_elapsed else 0.0
        duration = (1.0 * partial_elapsed / self.event_count) if self.event_count else 0.0
        error_rate = (1.0 * self.total_error_count / self.total_event_count) if self.total_event_count else 0.0
        success = 1.0 - error_rate
        partial_error_rate = (1.0 * self.error_count / self.event_count) if self.event_count else 0.0
        partial_success = 1.0 - partial_error_rate
        uptime_elapsed_td = now - self.run_ts
        uptime_elapsed = 1000.0 * uptime_elapsed_td.total_seconds()
        uptime = int(uptime_elapsed / 60000)
        stats = {
            'total_consumed_events': self.total_event_count,
            'total_errors': self.total_error_count,
            'avg_rate': avg_rate,
            'avg_event_duration': avg_duration,
            'avg_rate_ok_events': avg_rate_success,
            'avg_success': success,
            'avg_error_rate': error_rate,
            'elapsed_ms': int(partial_elapsed),
            'consumed_events': self.event_count,
            'errors': self.error_count,
            'rate': rate,
            'event_duration': duration,
            'rate_ok_events': rate_success,
            'uptime_minutes': uptime,
            'success_rate': partial_success,
            'error_rate': partial_error_rate
        }
        self.reset_batch(now)
        return stats
